import Ansi from './ansi.ts';

import { SCROLL_TAB_SIZE } from './config.ts';
import {
	arrayInsert,
	isAsciiDigit,
	isSeparator,
	strChars,
	strlen,
} from './fns.ts';
import { FileType } from './filetype/mod.ts';
import { highlightToColor, HighlightType } from './highlight.ts';
import Option, { None, Some } from './option.ts';
import { SearchDirection } from './types.ts';

/**
 * One row of text in the current document. In order to handle
 * multi-byte graphemes, all operations are done on an
 * array of 'character' strings.
 */
export class Row {
	/**
	 * The actual characters in the current row
	 */
	public chars: string[] = [];

	/**
	 * The characters rendered for the current row
	 * (like replacing tabs with spaces)
	 */
	public rchars: string[] = [];

	/**
	 * The syntax highlighting map
	 */
	public hl: HighlightType[] = [];

	private constructor(s: string | string[] = '') {
		this.chars = Array.isArray(s) ? s : strChars(s);
		this.rchars = [];
	}

	public get size(): number {
		return this.chars.length;
	}

	public get rsize(): number {
		return this.rchars.length;
	}

	public rstring(offset: number = 0): string {
		return this.rchars.slice(offset).join('');
	}

	public static default(): Row {
		return new Row();
	}

	public static from(s: string | string[] | Row): Row {
		if (s instanceof Row) {
			return s;
		}

		return new Row(s);
	}

	public append(s: string, syntax: FileType): void {
		this.chars = this.chars.concat(strChars(s));
		this.update(None, syntax);
	}

	public insertChar(at: number, c: string): void {
		const newSlice = strChars(c);
		if (at >= this.size) {
			this.chars = this.chars.concat(newSlice);
		} else {
			this.chars = arrayInsert(this.chars, at + 1, newSlice);
		}
	}

	/**
	 * Truncate the current row, and return a new one at the specified index
	 */
	public split(at: number, syntax: FileType): Row {
		const newRow = new Row(this.chars.slice(at));
		this.chars = this.chars.slice(0, at);
		this.update(None, syntax);

		return newRow;
	}

	/**
	 * Remove a character at the specified index
	 */
	public delete(at: number): void {
		if (at >= this.size) {
			return;
		}

		this.chars.splice(at, 1);
	}

	/**
	 * Search the current row for the specified string, and return
	 * the 'character' index of the start of that match
	 */
	public find(
		s: string,
		at: number = 0,
		direction: SearchDirection = SearchDirection.Forward,
	): Option<number> {
		if (at > this.size) {
			return None;
		}
		const thisStr = this.chars.join('');

		// Look for the search query `s`, starting from the 'character' `offset`
		const byteIndex = (direction === SearchDirection.Forward)
			? thisStr.indexOf(s, this.charIndexToByteIndex(at))
			: thisStr.lastIndexOf(s, this.charIndexToByteIndex(at));

		// No match after the specified offset
		if (byteIndex < 0) {
			return None;
		}

		// In many cases, the string length will
		// equal the number of characters. So
		// searching is fairly easy
		if (thisStr.length === this.chars.length) {
			return Some(this.cxToRx(byteIndex));
		}

		// Emoji/Extended Unicode-friendly search
		return Some(this.cxToRx(this.byteIndexToCharIndex(byteIndex)));
	}

	/**
	 * Convert the raw row offset to the equivalent offset for screen rendering
	 */
	public cxToRx(cx: number): number {
		let rx = 0;
		let j;
		for (j = 0; j < cx; j++) {
			if (this.chars[j] === '\t') {
				rx += (SCROLL_TAB_SIZE - 1) - (rx % SCROLL_TAB_SIZE);
			}
			rx++;
		}

		return rx;
	}

	/**
	 * Convert the screen rendering row offset to the file row offset
	 */
	public rxToCx(rx: number): number {
		let curRx = 0;
		let cx = 0;
		for (; cx < this.size; cx++) {
			if (this.chars[cx] === '\t') {
				curRx += (SCROLL_TAB_SIZE - 1) - (curRx % SCROLL_TAB_SIZE);
			}
			curRx++;

			if (curRx > rx) {
				return cx;
			}
		}

		return cx;
	}

	/**
	 * Convert the index of a JS string into the equivalent
	 * 'unicode character' index
	 */
	public byteIndexToCharIndex(byteIndex: number): number {
		if (this.toString().length === this.chars.length) {
			return byteIndex;
		}

		let n = 0;
		let byteCount = 0;
		for (; n < this.chars.length; n++) {
			byteCount += this.chars[n].length;
			if (byteCount > byteIndex) {
				return n;
			}
		}

		return this.chars.length;
	}

	/**
	 * Convert the 'unicode character' index into the equivalent
	 * JS string index
	 */
	public charIndexToByteIndex(charIndex: number): number {
		if (charIndex === 0 || this.toString().length === this.chars.length) {
			return charIndex;
		}

		// The char index will be the same size or smaller than
		// the JS string index, as a 'character' can consist
		// of multiple JS string indicies
		return this.chars.slice(0, charIndex).reduce(
			(prev, current) => prev + current.length,
			0,
		);
	}

	public toString(): string {
		return this.chars.join('');
	}

	public update(word: Option<string>, syntax: FileType): void {
		const newString = this.chars.join('').replaceAll(
			'\t',
			' '.repeat(SCROLL_TAB_SIZE),
		);

		this.rchars = strChars(newString);
		this.highlight(word, syntax);
	}

	public highlight(word: Option<string>, syntax: FileType): void {
		const highlighting = [];
		let searchIndex = 0;
		const matches = [];

		// Find matches for the current search
		if (word.isSome()) {
			while (true) {
				const match = this.find(word.unwrap(), searchIndex);
				if (match.isNone()) {
					break;
				}

				matches.push(match.unwrap());
				const nextPossible = match.unwrap() + strlen(word.unwrap());
				if (nextPossible < this.rsize) {
					searchIndex = nextPossible;
				} else {
					break;
				}
			}
		}

		let prevIsSeparator = true;
		let inString: string | boolean = false;
		let i = 0;
		for (; i < this.rsize;) {
			const ch = this.rchars[i];
			const prevHighlight = (i > 0) ? highlighting[i - 1] : HighlightType.None;

			// Highlight search matches
			if (word.isSome()) {
				if (matches.includes(i)) {
					for (const _ in strChars(word.unwrap())) {
						i += 1;
						highlighting.push(HighlightType.Match);
					}

					continue;
				}
			}

			// Highlight strings
			if (syntax.flags.strings) {
				if (inString) {
					highlighting.push(HighlightType.String);
					if (ch === inString) {
						inString = false;
					}
					i += 1;
					prevIsSeparator = true;
					continue;
				} else if (prevIsSeparator && ch === '"' || ch === "'") {
					highlighting.push(HighlightType.String);
					inString = ch;
					prevIsSeparator = true;
					i += 1;
					continue;
				}
			}

			// Highlight numbers
			if (syntax.flags.numbers) {
				const isNumeric = isAsciiDigit(ch) && (prevIsSeparator ||
					prevHighlight === HighlightType.Number);
				const isDecimalNumeric = ch === '.' &&
					prevHighlight === HighlightType.Number;
				const isHexNumeric = ch === 'x' &&
					prevHighlight === HighlightType.Number;
				if (isNumeric || isDecimalNumeric || isHexNumeric) {
					highlighting.push(HighlightType.Number);
				} else {
					highlighting.push(HighlightType.None);
				}
			}

			prevIsSeparator = isSeparator(ch);
			i += 1;
		}

		this.hl = highlighting;
	}

	public render(offset: number, len: number): string {
		const end = Math.min(len, this.rsize);
		const start = Math.min(offset, len);
		let result = '';
		let currentHighlight = HighlightType.None;

		for (let i = start; i < end; i++) {
			const highlightType = this.hl[i];

			if (highlightType !== currentHighlight) {
				currentHighlight = highlightType;
				result += highlightToColor(highlightType);
			}

			result += this.rchars[i];
		}

		result += Ansi.ResetFormatting;

		return result;
	}
}

export default Row;
